package qrz

import (
	"fmt"
	"io/ioutil"
	"log"
	"net/http"
	"regexp"
	"strings"

	"git.paulbsd.com/paulbsd/qrz/src/config"
	"github.com/antchfx/htmlquery"
	"github.com/robfig/cron"
)

// InitCronConfig create task schedules
func InitCronConfig(config config.Config) (err error) {
	cr := cron.New()
	cr.AddFunc(config.Cron, func() { err = Run(config) })
	cr.Start()
	return
}

// Run runs the main loop
func Run(config config.Config) (err error) {
	for _, group := range config.QrzGroups {
		url := fmt.Sprintf(config.URLBase, group)
		fmt.Println(fmt.Sprintf("Processing the %s group", group))

		bodyStr, err := getBody(url)
		if err != nil {
			log.Fatalln(err)
		}

		frsPeople, err := getFrsEntries(config, bodyStr)
		if err != nil {
			log.Fatalln(err)
		}

		existingQrz, err := getCurrentEntries(config)
		if err != nil {
			log.Fatalln(err)
		}

		err = discardExistingEntries(config, &frsPeople, existingQrz)
		if err != nil {
			log.Fatalln(err)
		}

		if len(frsPeople) > 0 {
			err = insertFrsEntryToDB(config, frsPeople)
			if err != nil {
				log.Println(err)
			}
		} else {
			fmt.Println(
				fmt.Sprintf("No entry to insert in database for group %s",
					group))
		}
	}
	return
}

// getBody fetch html page
func getBody(url string) (string, error) {
	var client http.Client

	resp, err := client.Get(url)
	if err != nil {
		return "", err
	}

	body, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		return "", err
	}
	defer resp.Body.Close()
	/*
		// For old website support 23/05/2020
		dec := charmap.ISO8859_1.NewDecoder()
		output, _ := dec.Bytes(body)
		bodyString := string(output)
	*/
	return string(body), nil
}

// getFrsEntries get FRS entries from html body
func getFrsEntries(config config.Config, body string) (frsentries map[string]Qrz, err error) {
	frsentries = make(map[string]Qrz)
	re := regexp.MustCompile(`^   [0-9]{1,4}\s[A-Z]{1,4}\s[0-9]{1,4}`)

	htmlpage, err := htmlquery.Parse(strings.NewReader(body))
	if err != nil {
		return
	}

	for _, n := range htmlquery.Find(htmlpage, "//tr") {
		td := htmlquery.Find(n, "//td")
		if len(td) > 2 {
			if re.MatchString(htmlquery.InnerText(td[1])) {
				frs := Qrz{
					QRZ:     strings.TrimLeft(htmlquery.InnerText(td[1]), " "),
					DMRID:   strings.TrimLeft(htmlquery.InnerText(td[0]), " "),
					Name:    strings.TrimLeft(htmlquery.InnerText(td[2]), " "),
					City:    strings.TrimLeft(htmlquery.InnerText(td[3]), " "),
					Dept:    strings.TrimLeft(htmlquery.InnerText(td[4]), " "),
					Country: strings.TrimLeft(htmlquery.InnerText(td[5]), " ")}
				frsentries[frs.QRZ] = frs
			} else if re.MatchString(htmlquery.InnerText(td[0])) {
				frs := Qrz{
					QRZ:     strings.TrimLeft(htmlquery.InnerText(td[0]), " "),
					Name:    strings.TrimLeft(htmlquery.InnerText(td[1]), " "),
					City:    strings.TrimLeft(htmlquery.InnerText(td[2]), " "),
					Dept:    strings.TrimLeft(htmlquery.InnerText(td[3]), " "),
					Country: strings.TrimLeft(htmlquery.InnerText(td[4]), " ")}
				frsentries[frs.QRZ] = frs
			}
		}
	}

	return
}

// getCurrentEntries fetch existing entries from database
func getCurrentEntries(config config.Config) (qrz []Qrz, err error) {
	err = config.Db.Find(&qrz)

	return
}

// discardExistingEntries remove existing entries from original map[string]Qrz
func discardExistingEntries(config config.Config, frsPeople *map[string]Qrz, existingQRZ []Qrz) (err error) {
	for _, entry := range existingQRZ {
		delete(*frsPeople, entry.QRZ)
	}

	return
}

// insertFrsEntryToDB inserts frs entries to database
func insertFrsEntryToDB(config config.Config, frsPeople map[string]Qrz) (err error) {
	session := config.Db.NewSession()
	defer session.Close()

	err = session.Begin()

	var qrzNum int

	fmt.Println(
		fmt.Sprintf("Starting insert of %d entries",
			len(frsPeople)))

	for _, frs := range frsPeople {
		_, err = session.Insert(&frs)
		qrzNum++
		if (qrzNum % 5) == 0 {
			err = session.Commit()
			if err != nil {
				return err
			}
		}
	}

	err = session.Commit()
	if err != nil {
		return err
	}
	fmt.Println(fmt.Sprintf("Committed %d qrz entries", qrzNum))

	return nil
}

// ToSlice converts struct to array
/* func ToSlice(qrz Qrz) (out []interface{}) {
	val := reflect.ValueOf(qrz)
	num := val.NumField()

	for i := 1; i < num-1; i++ {
		out = append(out, val.Field(i).Interface())
	}
	return
} */

// ToSlice converts struct to array
func ToSlice(qrz Qrz) (out []string) {
	out = []string{
		qrz.QRZ,
		qrz.Name,
		qrz.City,
		qrz.Dept,
		qrz.Country,
		qrz.DMRID}
	return
}

// Qrz describe FRS people
type Qrz struct {
	ID      int    `db:"id" xorm:"pk autoincr"`
	QRZ     string `db:"qrz" xorm:"varchar(25) notnull"`
	Name    string `db:"name" xorm:"varchar(25) notnull"`
	Address string `db:"address" xorm:"varchar(50) notnull"`
	City    string `db:"city" xorm:"varchar(50) notnull"`
	Zipcode string `db:"zipcode" xorm:"varchar(5) notnull"`
	Dept    string `db:"dept" xorm:"varchar(25) notnull"`
	Country string `db:"country" xorm:"varchar(25) notnull"`
	DMRID   string `db:"dmrid" xorm:"varchar(25) notnull"`
}
